from json import dump, load
from pathlib import Path
from typing import Optional


class FileAdapter:
    def __init__(self, path: Path):
        self.path = path

    def create(self, data: Optional[dict] = None):
        if not self.path.parent.exists():
            self.path.parent.mkdir(parents=True)

        if not self.path.exists():
            self.save(data or {})

            return True

        return False

    def delete(self):
        self.path.unlink()

    def load(self) -> dict:
        with open(self.path) as fd:
            return load(fd)

    def save(self, data: dict) -> None:
        with open(self.path, "w") as fd:
            dump(data, fd, indent=2)



from os import environ
from pathlib import Path
from typing import Optional

from pydantic import BaseModel

from macro_counter.adapters.file import FileAdapter

DEFAULT_CONFIG_PATH = Path.home() / ".config" / "macro_counter" / "config.json"
DEFAULT_STORE_PATH = Path.home() / ".config" / "macro_counter" / "store.json"


class MongoSettings(BaseModel):
    username: Optional[str]
    password: Optional[str]
    database: Optional[str]
    host: Optional[str]
    port: Optional[int] = 27017
    srv_mode: Optional[bool] = False
    timeout_ms: Optional[int] = 2000

    @property
    def is_valid(self):
        return all([self.username, self.password, self.database, self.host])


class EnvSettings(BaseModel):
    config_path: Optional[str] = environ.get("MACRO_COUNTER_CONFIG_PATH")
    local_store_path: Optional[str] = environ.get("MACRO_COUNTER_LOCAL_STORE_PATH")

    username: Optional[str] = environ.get("MACRO_COUNTER_MONGODB_USERNAME")
    password: Optional[str] = environ.get("MACRO_COUNTER_MONGODB_PASSWORD")
    database: Optional[str] = environ.get("MACRO_COUNTER_MONGODB_DATABASE")
    host: Optional[str] = environ.get("MACRO_COUNTER_MONGODB_HOST")
    port: Optional[str] = environ.get("MACRO_COUNTER_MONGODB_PORT")
    srv_mode: Optional[str] = environ.get("MACRO_COUNTER_MONGODB_SRV_MODE")
    timeout_ms: Optional[str] = environ.get("MACRO_COUNTER_MONGODB_TIMEOUT_MS")


class FileSettings(BaseModel):
    mongo_settings: MongoSettings = MongoSettings()

    local_store_path: Optional[str]


class AppSettings(BaseModel):
    env_settings: EnvSettings
    file_settings: FileSettings

    config_path: Optional[Path] = None
    mongo_settings: Optional[MongoSettings] = None
    local_store_path: Path = DEFAULT_STORE_PATH


def get_settings():
    """
    Build a setting object containing both file configuration and env variables
    """

    env_settings = EnvSettings()

    if (
        env_settings.config_path
        and (config_path := Path(env_settings.config_path)).exists()
    ):
        config_file = FileAdapter(config_path)
    else:
        config_path = DEFAULT_CONFIG_PATH

        config_file = FileAdapter(config_path)

        if not config_path.exists():
            print(f"Empty setting file created: {config_path}")

            config_file.create(FileSettings().dict())

    file_settings = FileSettings(**config_file.load())

    app_settings = AppSettings(
        env_settings=env_settings,
        file_settings=file_settings,
        config_path=str(config_file.path),
    )

    if file_settings.local_store_path:
        app_settings.local_store_path = Path(file_settings.local_store_path)

    app_settings.mongo_settings = MongoSettings(
        **{
            **file_settings.mongo_settings.dict(exclude_none=True),
            **env_settings.dict(exclude_none=True),
        }
    )

    return app_settings



from typing import List

from macro_counter.adapters import FileAdapter
from macro_counter.exceptions import ComponentAlreadyExist, ComponentDoesNotExist
from macro_counter.models import Component
from macro_counter.repositories.components.abstract import AbstractComponentRepository


class LocalComponentRepository(AbstractComponentRepository):
    def __init__(self, file_adapter: FileAdapter):
        self.file = file_adapter

    def create(self, component: Component) -> Component:
        local_data = self.file.load()

        if local_data.get(component.name):
            raise ComponentAlreadyExist(component.name)
        else:
            local_data[component.name] = component.dict()

            self.file.save(local_data)

            return component

    def update(self, component: Component) -> Component:
        local_component = self.get(component.name)

        if local_component.dict() != component.dict():
            local_data = self.file.load()

            local_data[component.name] = component.dict()

            self.file.save(local_data)

        return component

    def get(self, component_name: str) -> Component:
        local_data = self.file.load()

        if component_data := local_data.get(component_name):
            return Component(**component_data)
        else:
            raise ComponentDoesNotExist(component_name)

    def list(self) -> List[Component]:
        return [
            Component(**component_data) for component_data in self.file.load().values()
        ]

    def delete(self, component: Component) -> bool:
        local_data = self.file.load()

        if local_data.get(component.name):
            del local_data[component.name]

            self.file.save(local_data)

            return True
        else:
            raise ComponentDoesNotExist(component.name)

    def delete_all(self) -> bool:
        self.file.save({})

        return True



from pyparsing import (
    CaselessLiteral,
    Empty,
    Group,
    Optional,
    Suppress,
    Word,
    ZeroOrMore,
    alphas,
    delimitedList,
    nums,
)

integer = Word(nums)
float_num = Word(nums + ".")
component = Word(alphas + nums + "_")("component")

"""
COMPONENT_PARSER = Word(alphas + nums + "_")("component")

VALIDATED_COMPONENT_PARSER = (
    COMPONENT_PARSER
    #.setParseAction(validate_component)
)
"""

normalization = Suppress("%") + float_num("normalization").setParseAction(
    lambda *args: float(args[2][0])
)

operations = Group(
    (
        CaselessLiteral("*").setParseAction(lambda: "multiply")
        | CaselessLiteral("/").setParseAction(lambda: "divide")
    )("op")
    + float_num("number").setParseAction(lambda *args: float(args[2][0]))
).setResultsName("operations", listAllMatches=True)

modifiers = ZeroOrMore(normalization | operations)

component_group = delimitedList(
    Group(ZeroOrMore(component("name") + Optional(modifiers))), delim="+"
)("components")

app_parser = (
    Suppress("register")("register").setParseAction(lambda: True) + Optional(component)
    | Suppress("delete")("delete").setParseAction(lambda: True) + Optional(component)
    | component("assign") + Suppress("=") + component_group
    | Suppress("detail")("detail").setParseAction(lambda: True) + component_group
    | Empty()("display").setParseAction(lambda: True) + Optional(component_group)
)



from abc import ABC, abstractmethod
from typing import List

from macro_counter.models import Component


class AbstractComponentRepository(ABC):
    @abstractmethod
    def create(self, component: Component) -> Component:
        pass

    @abstractmethod
    def update(self, component: Component) -> Component:
        pass

    @abstractmethod
    def get(self, component_name: str) -> Component:
        pass

    @abstractmethod
    def list(self) -> List[Component]:
        pass

    @abstractmethod
    def delete(self, component: Component) -> bool:
        pass

    @abstractmethod
    def delete_all(self) -> bool:
        pass



from typing import Type, Union

from macro_counter.adapters import AdapterInstance, MongoAdapter
from macro_counter.repositories.components.local import LocalComponentRepository
from macro_counter.repositories.components.mongo import MongoComponentRepository

ComponentRepository = Union[
    Type[LocalComponentRepository],
    Type[MongoComponentRepository],
]


def component_repository_factory(adapter: AdapterInstance) -> ComponentRepository:
    if isinstance(adapter, MongoAdapter):
        return MongoComponentRepository
    else:
        return LocalComponentRepository



from typing import List

from macro_counter.adapters import MongoAdapter
from macro_counter.exceptions import ComponentAlreadyExist, ComponentDoesNotExist
from macro_counter.models import Component
from macro_counter.repositories.components.abstract import AbstractComponentRepository


class MongoComponentRepository(AbstractComponentRepository):
    def __init__(self, mongo_adapter: MongoAdapter, collection_name="components"):
        self.mongo_collection = mongo_adapter.get_collection(collection_name)

    def create(self, component: Component) -> Component:
        if self.mongo_collection.find_one({"name": component.name}):
            raise ComponentAlreadyExist(component.name)
        else:
            self.mongo_collection.insert_one(component.dict())

            return component

    def update(self, component: Component) -> Component:
        mongo_component = self.get(component.name)

        if mongo_component.dict() != component.dict():
            self.mongo_collection.update_one(
                {"name": component.name}, {"$set": component.dict()}
            )

        return component

    def get(self, component_name: str) -> Component:
        if component_data := self.mongo_collection.find_one({"name": component_name}):
            return Component(**component_data)
        else:
            raise ComponentDoesNotExist(component_name)

    def list(self) -> List[Component]:
        return [
            Component(**component_data)
            for component_data in self.mongo_collection.find()
        ]

    def delete(self, component: Component) -> bool:
        if self.mongo_collection.find_one({"name": component.name}):
            self.mongo_collection.delete_one({"name": component.name})

            return True
        else:
            raise ComponentDoesNotExist(component.name)

    def delete_all(self) -> bool:
        for component in self.list():
            self.delete(component)

        return True



from macro_counter.models import Field

unit_field = Field(label="units", fullname="Units")

protein_field = Field(
    label="protein",
    fullname="Protein",
    shortname="Prot",
    macro=True,
    show_percents=True,
)
carb_field = Field(label="carb", fullname="Carb", macro=True, show_percents=True)
fat_field = Field(label="fat", fullname="Fat", macro=True, show_percents=True)

macro_fields = [protein_field, carb_field, fat_field]

attrs_fields = [
    Field(label="kcal", fullname="Calories", shortname="Cal"),
    protein_field,
    carb_field,
    Field(
        label="fiber",
        fullname="Fiber",
    ),
    Field(label="sugar", fullname="Sugar", show_percents=True),
    fat_field,
    Field(
        label="saturated_fat",
        fullname="Saturated fat",
        shortname="Sat",
        show_percents=True,
    ),
    Field(
        label="mono_fat",
        fullname="Mono insaturated fat",
        shortname="Mono",
        show_percents=True,
    ),
    Field(
        label="poly_fat",
        fullname="Poly insaturated fat",
        shortname="Poly",
        show_percents=True,
    ),
    Field(
        label="trans_fat", fullname="Trans fat", shortname="Trans", show_percents=True
    ),
]



from copy import deepcopy
from enum import Enum, auto

from prompt_toolkit import print_formatted_text, prompt
from prompt_toolkit.completion import WordCompleter
from prompt_toolkit.history import FileHistory
from pyparsing import ParseFatalException
from tabulate import tabulate

from macro_counter.adapters import get_adapters
from macro_counter.app.parsers import app_parser
from macro_counter.exceptions import ComponentDoesNotExist
from macro_counter.fields import attrs_fields, macro_fields, unit_field
from macro_counter.models import Component, ComponentKind
from macro_counter.repositories.components import component_repository_factory
from macro_counter.utils import is_float


class PromptSignal(Enum):
    LEAVE = auto()
    QUIT = auto()


class PromptState(Enum):
    RUNNING = auto()
    STOPPED = auto()


EmptyWordCompleter = WordCompleter([])


def get_measure(component: Component):
    if component.kind == ComponentKind.SOLID:
        return "gr"
    else:
        return "ml"


class AppPrompt:
    def __init__(self):
        self.adapters = get_adapters()
        self.repo = component_repository_factory(self.adapters.current)(
            self.adapters.current
        )
        self.state = PromptState.STOPPED
        self.history = FileHistory(".macro_counter_history")

        self.completer = None
        self.reset_completer()

    def prompt(self, message, completer=None, use_history=True):
        return prompt(
            message=message,
            history=self.history if use_history else None,
            completer=self.completer if completer is None else completer,
            enable_suspend=True,
        )

    def reset_completer(self):
        self.components = {component.name: component for component in self.repo.list()}
        self.completer = WordCompleter(
            [
                "+",
                "%",
                "*",
                "/",
                "=",
                "register",
                "delete",
                "detail",
                "quit",
                *sorted(self.components.keys()),
            ]
        )

    def print(self, *args, **kwargs):
        print_formatted_text(*args, **kwargs)

    def _get_kind(self, default=None):
        kind_raw = self.prompt(
            f"Type (L)iquid/(S)olid{' (' + str(default.capitalize()) + ') ' if default else ''}: ",
            completer=WordCompleter(["Solid", "Liquid"], ignore_case=True),
            use_history=False,
        )

        if not kind_raw and default:
            kind = default
        elif kind_raw.lower() in ("l", "liquid"):
            kind = "liquid"
        elif kind_raw.lower() in ("s", "solid"):
            kind = "solid"
        else:
            raise ValueError("Error: Wrong kind")

        return kind

    def _get_units(self, default=None):
        units_raw = self.prompt(
            f"Type units{' (' + str(default) + ') ' if default else ''}: ",
            use_history=False,
            completer=EmptyWordCompleter,
        )

        if not units_raw:
            units = default
        elif is_float(units_raw):
            units = float(units_raw)
        else:
            raise ValueError("Error: Wrong value, number required")

        return units

    def _get_attrs(self, defaults=None):
        attrs = deepcopy(defaults) or {}

        for field in attrs_fields:
            value = None

            if default := attrs.get(field.label):
                value = self.prompt(
                    f"How much {field.fullname} ({default}/Reset): ",
                    completer=WordCompleter(["Reset"], ignore_case=True),
                    use_history=False,
                )
            else:
                value = self.prompt(
                    f"How much {field.fullname} : ",
                    completer=EmptyWordCompleter,
                    use_history=False,
                )

            if value.lower() in ("r", "reset"):
                attrs.pop(field.label)
                value = None
            elif value:
                if is_float(value):
                    value = float(eval(value))
                else:
                    raise ValueError("Error: Wrong value, number required")
            elif default:
                value = float(default)

            if value:
                attrs[field.label] = value

        return attrs

    def create(self, name):
        self.print(f"Registering {name}")

        try:
            kind = self._get_kind()
            units = self._get_units(default=100.0)
            attrs = self._get_attrs()

        except Exception as exc:
            self.print(str(exc))

            return

        self.repo.create(Component(name=name, kind=kind, units=units, attrs=attrs))

        self.print(f"Created: {name}")
        self.reset_completer()

    def update(self, component):
        self.print(f"Updating {component.name}")

        altered = False

        try:
            kind = self._get_kind(default=component.kind)

            if kind != component.kind:
                component.kind = kind
                altered = True

            units = self._get_units(default=component.units)

            if units != component.units:
                component.units = units
                altered = True

            attrs = self._get_attrs(defaults=component.attrs)

            if attrs != component.attrs:
                component.attrs = attrs
                altered = True

        except Exception as exc:
            self.print(str(exc))

            return

        if altered:
            self.repo.update(component)

            self.print(f"Updated: {component.name}")
        else:
            self.print("Nothing changed")

    def register(self, name):
        try:
            component = self.repo.get(name)

            self.update(component)
        except ComponentDoesNotExist:
            self.create(name)

        self.reset_completer()

    def delete(self, name):
        try:
            component = self.repo.get(name)

            self.repo.delete(component)

            del self.components[name]

            self.reset_completer()
            self.print(f"Component {name} deleted")

        except ComponentDoesNotExist:
            self.print(f"Component {name} not found")

    def loop(self):
        self.state = PromptState.RUNNING

        while True:
            try:
                signal = self.loop_step()
                if signal == PromptSignal.LEAVE:
                    break
                if signal == PromptSignal.QUIT:
                    self.state = PromptState.STOPPED

                    return PromptSignal.QUIT
            except EOFError:
                self.print("EOF : quitting...")

                self.state = PromptState.STOPPED

                return PromptSignal.QUIT
            except KeyboardInterrupt:
                pass

        self.state = PromptState.STOPPED

    def loop_step(self):
        text = self.prompt(">>> ")

        if text in ("q", "quit"):
            self.print("quitting...")

            return PromptSignal.QUIT

        elif text in ("l", "leave"):
            self.print("leaving...")

            return PromptSignal.LEAVE

        elif text:
            try:
                if signal := self.dispatch(text):
                    return signal

            except ParseFatalException as exc:
                offset = len(self.session.message) + exc.loc

                self.print(f"{' ' * offset}^ {exc.msg}")

    def compute_components(self, parsed_components):
        components = []

        for parsed_component in parsed_components:
            component_data = parsed_component.asDict()

            if not component_data:
                continue

            component_name = component_data["name"]

            if not (component := self.components.get(component_name)):
                try:
                    component = self.repo.get(component_name)
                except ComponentDoesNotExist:
                    pass

            if not component:
                self.print(f"No component {component_name} has been found: skipping")

                continue

            if normalization := component_data.get("normalization"):
                component = component % normalization

            if operations := component_data.get("operations"):
                multiplier = 1

                for operation in operations:
                    if operation["op"] == "multiply":
                        multiplier *= operation["number"]

                    elif operation["op"] == "divide":
                        multiplier /= operation["number"]

                component = component * multiplier

            components.append(component)

        return components

    def display(self, component):
        macro_total = sum(
            [
                component.attrs[field.label]
                for field in macro_fields
                if field.label in component.attrs
            ]
        )

        rows = []

        for field in attrs_fields:
            value = component.attrs.get(field.label)

            if not value:
                continue

            field_label_cell = field.fullname
            if field.show_percents and not field.macro:
                field_label_cell = "- " + field_label_cell

            value_cell = f"{value:.1f}"

            percent_cell = None
            if field.show_percents:
                percent_cell = f"{value / macro_total * 100:.1f}%"

            rows.append([field_label_cell, value_cell, percent_cell])

        rows.insert(
            1,
            [
                unit_field.fullname,
                f"{component.units} {get_measure(component)}",
            ],
        )

        self.print(tabulate(rows))

    def display_details(self, component, components):
        rows = []

        present_fields = [
            field for field in attrs_fields if field.label in component.attrs
        ]
        present_fields.insert(0, unit_field)

        headers = ["Name"] + [field.name for field in present_fields]

        total = {field.label: 0.0 for field in present_fields}

        for component in components:
            total[unit_field.label] += component.units
            for k, v in component.attrs.items():
                total[k] = total[k] + v

        def to_percent(num, total, vv=1):
            return f"{num / (total if total else 1) * 100:.{vv}f}%"

        for component in components:
            raw_data = [component.name]
            perc_data = [None]

            for field in present_fields:
                if field is unit_field:
                    raw_data.append(str(component.units) + get_measure(component))
                    perc_data.append(
                        to_percent(component.units, total[unit_field.label])
                    )
                elif value := component.attrs.get(field.label):
                    raw_data.append(value)
                    perc_data.append(to_percent(value, total[field.label]))
                else:
                    raw_data.append(None)
                    perc_data.append(None)

            rows.append(raw_data)
            rows.append(perc_data)

        rows.append([])

        row = ["Total"]
        for field in present_fields:
            if field is unit_field:
                row.append(total[unit_field.label])
            else:
                row.append(f"{total[field.label]:.1f}")

        rows.append(row)

        self.print(tabulate(rows, headers=headers))

    def dispatch(self, text):
        parsed = app_parser.parseString(text)

        if parsed.register:
            if name := parsed.component:
                self.register(name)
            else:
                self.print("Component name must be set")

        elif parsed.delete:
            if name := parsed.component:
                self.delete(name)
            else:
                self.print("Component name must be set")
        else:
            components = self.compute_components(parsed.components)

            if not components:
                self.print("At least one component must be set")

                return

            component = sum(components[1:], start=deepcopy(components)[0])
            component.name = parsed.assign

            if parsed.assign:
                component.kind = self._get_kind(default=component.kind)
                component.units = self._get_units(default=component.units)

                try:
                    self.repo.update(component)

                    self.print(f"Updated: {component.name}")
                except ComponentDoesNotExist:
                    self.repo.create(component)

                    self.print(f"Created: {component.name}")
                    self.reset_completer()

            elif parsed.display:
                self.display(component)

            elif parsed.detail:
                self.display_details(component, components)



from typing import Optional, Union

from pydantic import BaseModel

from macro_counter.adapters.file import FileAdapter
from macro_counter.adapters.mongo import MongoAdapter
from macro_counter.core.settings import get_settings

AdapterInstance = Union[FileAdapter, MongoAdapter]


class Adapters(BaseModel):
    file: FileAdapter
    mongo: Optional[MongoAdapter]
    current: AdapterInstance

    class Config:
        arbitrary_types_allowed = True


def get_adapters() -> Adapters:
    settings = get_settings()

    mongo = (
        MongoAdapter(settings.mongo_settings)
        if settings.mongo_settings.is_valid
        else None
    )
    file = FileAdapter(settings.local_store_path)
    file.create()

    return Adapters(
        mongo=mongo, file=file, current=mongo if mongo and mongo.connected else file
    )



from pymongo import MongoClient
from pymongo.errors import ServerSelectionTimeoutError


class MongoAdapter:
    def __init__(self, config):
        self.connect(config)

    def connect(self, config):
        self._client = MongoClient(
            self.format_uri(config), serverSelectionTimeoutMS=config.timeout_ms
        )
        self.database = getattr(self._client, config.database)

    @property
    def connected(self):
        try:
            self._client.server_info()
            return True
        except ServerSelectionTimeoutError:
            return False

    def get_collection(self, collection_name: str):
        return getattr(self.database, collection_name)

    def format_uri(self, config):
        if config.srv_mode:
            uri = "mongodb+srv://{username}:{password}@{host}/{database}?retryWrites=true&w=majority"
        else:
            uri = "mongodb://{username}:{password}@{host}:{port}/{database}?authSource=admin"

        return uri.format(**config.dict())



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['macro_counter',
 'macro_counter.adapters',
 'macro_counter.app',
 'macro_counter.core',
 'macro_counter.repositories.components']

package_data = \
{'': ['*']}

install_requires = \
['dnspython>=2.1.0,<3.0.0',
 'prompt-toolkit>=3.0.18,<4.0.0',
 'pydantic>=1.9.0,<2.0.0',
 'pymongo>=3.11.4,<4.0.0',
 'pyparsing>=2.4.7,<3.0.0',
 'tabulate>=0.8.9,<0.9.0']

entry_points = \
{'console_scripts': ['macro_counter = macro_counter:main']}

setup_kwargs = {
    'name': 'macro-counter',
    'version': '0.2.1',
    'description': 'Python macro nutrient counter.',
    'long_description': "# Macro Counter\n\nConvenient terminal application to keep track of calories/macros with a simple prompt\n\n## Installation\n\nMacro counter can be installed from test PyPI using a little bit customized `pip` command:\n\n```\npip3 install --upgrade -i https://test.pypi.org/simple/ --extra-index-url https://pypi.python.org/simple macro_counter\n```\n\n## Usage\n\n### CLI\n\n**Register a component**\n\n```\n>>> register tomato_100gr\nRegistering tomato_100gr\nType (L)iquid/(S)olid: Solid\nType units (100.0) :\nHow much Calories : 18\nHow much Protein : 0.9\nHow much Carb : 3.9\nHow much Fiber : 1.2\nHow much Sugar : 2.6\nHow much Fat : 0.2\nHow much Saturated fat :\nHow much Mono insaturated fat :\nHow much Poly insaturated fat : 0.1\nHow much Trans fat :\nCreated: tomato_100gr\n>>> register coco_milk_100ml\nRegistering coco_milk_100ml\nType (L)iquid/(S)olid: Liquid\nType units (100.0) :\nHow much Calories : 185\nHow much Protein : 1.6\nHow much Carb : 2\nHow much Fiber :\nHow much Sugar : 2\nHow much Fat : 19\nHow much Saturated fat : 17\nHow much Mono insaturated fat :\nHow much Poly insaturated fat :\nHow much Trans fat :\nCreated: coco_milk_100ml\n```\n\n**Checking component infos**\n\n```\n>>> tomato_100gr\n----------------------  --------  -----\nCalories                18.0\nUnits                   100.0 gr\nProtein                 0.9       18.0%\nCarb                    3.9       78.0%\nFiber                   1.2\n- Sugar                 2.6       52.0%\nFat                     0.2       4.0%\n- Poly insaturated fat  0.1       2.0%\n----------------------  --------  -----\n>>> coco_milk_100ml\n---------------  --------  -----\nCalories         185.0\nUnits            100.0 ml\nProtein          1.6       7.1%\nCarb             2.0       8.8%\n- Sugar          2.0       8.8%\nFat              19.0      84.1%\n- Saturated fat  17.0      75.2%\n---------------  --------  -----\n```\n\n**Multiplying operations**\n\nTo check the nutritional facts of 2 liters of Coco milk.\n\n```\n>>> coco_milk_100ml * 20\n---------------  ---------  -----\nCalories         3700.0\nUnits            2000.0 ml\nProtein          32.0       7.1%\nCarb             40.0       8.8%\n- Sugar          40.0       8.8%\nFat              380.0      84.1%\n- Saturated fat  340.0      75.2%\n---------------  ---------  -----\n```\n\nTo check the nutritional facts of 2 liters of Coco milk using normalizing-to-one operation.\n\n```\n>>> coco_milk_100ml % 2000\n---------------  ---------  -----\nCalories         3700.0\nUnits            2000.0 ml\nProtein          32.0       7.1%\nCarb             40.0       8.8%\n- Sugar          40.0       8.8%\nFat              380.0      84.1%\n- Saturated fat  340.0      75.2%\n---------------  ---------  -----\n```\n\n**Adding operations**\n\n```\n>>> tomato_100gr + coco_milk_100ml\n----------------------  --------  -----\nCalories                203.0\nUnits                   200.0 gr\nProtein                 2.5       9.1%\nCarb                    5.9       21.4%\nFiber                   1.2\n- Sugar                 4.6       16.7%\nFat                     19.2      69.6%\n- Saturated fat         17.0      61.6%\n- Poly insaturated fat  0.1       0.4%\n----------------------  --------  -----\n```\n\n**Updating existing components**\n\nYou can remove actual fields using 'r' or 'reset' keyword.\n\n```\n>>> register tomato_100gr\nUpdating tomato_100gr\nType (L)iquid/(S)olid (Solid) :\nType units (100.0) :\nHow much Calories (18.0/Reset):\nHow much Protein (0.9/Reset):\nHow much Carb (3.9/Reset):\nHow much Fiber (1.2/Reset):\nHow much Sugar (2.6/Reset):\nHow much Fat (0.2/Reset):\nHow much Saturated fat :\nHow much Mono insaturated fat :\nHow much Poly insaturated fat (0.1/Reset):\nHow much Trans fat :\nNothing changed\n```\n\n**Assign a single component**\n\nYou can also update the unit field, for example cooked chicken won't be as heavy as the raw one, but will still contains the macros.\n\n```\n>>> chicken_cooked = chicken_raw\nType (L)iquid/(S)olid (Solid) :\nType units (200.0) : 160\nCreated: chicken_cooked\n```\n\n**Assign a recipe**\n\nWeight the product at the end of the recipe to fine tune further macro counting, corresponding to weight gain according to cooking, evaporating water, ect...\n\n```\n>>> tiramisu = eggs * 4 + almond_flour % 66 + mascarpone % 500 + erythritol * 66 * 22 + fresh_cream % 200\nType (L)iquid/(S)olid (Solid) :\nType units (1000.0) : 900\nCreated: tiramisu\n```\n\n**Delete a component**\n\n```\n>>> delete tomato\nComponent tomato deleted\n```\n\n**Display components details**\n\nThis will display each components data with their percentages over all the other.\n\n```\n>>> detail tomato_100gr + coco_milk_100ml\nName             Units    Cal    Prot    Carb    Fiber    Sugar    Fat    Sat     Poly\n---------------  -------  -----  ------  ------  -------  -------  -----  ------  ------\ntomato_100gr     100.0gr  18.0   0.9     3.9     1.2      2.6      0.2            0.1\n                 50.0%    8.9%   36.0%   66.1%   100.0%   56.5%    1.0%           100.0%\ncoco_milk_100ml  100.0ml  185.0  1.6     2.0              2.0      19.0   17.0\n                 50.0%    91.1%  64.0%   33.9%            43.5%    99.0%  100.0%\n\nTotal            200.0    203.0  2.5     5.9     1.2      4.6      19.2   17.0    0.1\n```\n",
    'author': 'Laurent Arthur',
    'author_email': 'laurent.arthur75@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/ludwig778/macro_counter',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from typing import Optional

from pydantic import BaseModel
from pydantic import Field as PydanticField
from pydantic import validator


class ComponentKind:
    SOLID = "solid"
    LIQUID = "liquid"


class Component(BaseModel):
    name: str
    kind: str

    units: float = 1.0
    attrs: dict = PydanticField(default_factory=dict)

    @validator("kind")
    def validate_kind(cls, value):
        if value not in ("solid", "liquid"):
            raise ValueError('Component kind must be either "solid" or "liquid"')
        return value

    class Config:
        extra = "ignore"

    def copy(self):
        return Component(
            name=self.name, kind=self.kind, units=self.units, attrs=self.attrs
        )

    def multiply(self, val):
        self.units *= val

        new_attrs = {}

        for k, v in self.attrs.items():
            if v:
                new_attrs[k] = v * val

        self.attrs = new_attrs

    def __add__(self, other):
        obj = self.copy()

        obj.units += other.units

        for k, v in other.attrs.items():
            if k not in obj.attrs:
                obj.attrs[k] = 0

            obj.attrs[k] += v

        return obj

    def __mod__(self, val):
        obj = self.copy()

        obj.multiply(1 / obj.units)
        obj.multiply(val)

        return obj

    def __mul__(self, val):
        obj = self.copy()
        obj.multiply(val)

        return obj

    def __truediv__(self, val):
        obj = self.copy()
        obj.multiply(1 / val)

        return obj


class Field(BaseModel):
    label: str
    fullname: str
    shortname: Optional[str]
    macro: bool = False
    show_percents: bool = False

    @property
    def name(self):
        return self.shortname or self.fullname


"""
class Component:
    def __init__(self, name, kind=None, units=None, attrs=None, components=None, **kwargs):
        self.name = name

        if not kind:
            raise Exception("Kind must be set")

        self.kind = kind
        self.units = units or 1
        self.attrs = attrs or {}

        self.components = components or {}

    @classmethod
    def create(cls, name, **kwargs):
        if not (obj := cls.get(name)):
            obj = cls(name, **kwargs)

            component_collection.insert_one(obj.to_dict())

        return obj

    @property
    def measure(self):
        return "gr" if self.kind == "solid" else "ml"

    @classmethod
    def list(cls, **kwargs):
        return [
            cls(**ingr)
            for ingr in component_collection.find(kwargs)
        ]

    @classmethod
    def get(cls, name):
        component = component_collection.find_one({"name": name})

        if component:
            return cls(**component)

    def update(self, **kwargs):
        self_data = self.to_dict()
        new_data = {**self_data, **kwargs}

        if self_data != new_data:
            component_collection.update_one({"name": self.name}, {"$set": new_data})

            self.__dict__.update(new_data)

    def delete(self):
        return component_collection.delete_one({"name": self.name})

    def copy(self):
        return self.__class__(**self.to_dict())

    def to_dict(self):
        return {
            "name": self.name,
            "units": self.units,
            "kind": self.kind,
            "attrs": self.attrs,
            "components": self.components
        }

    def __repr__(self):
        return f"<{self.__class__.__name__}: {self.name} {self.units}{self.measure}>"

    def multiply(self, val):
        self.units *= val

        new_attrs = {}

        for k, v in self.attrs.items():
            if v:
                new_attrs[k] = v * val

        self.attrs = new_attrs

    def __mod__(self, val):
        obj = self.copy()

        obj.multiply(1 / obj.units)
        obj.multiply(val)

        return obj

    def __mul__(self, val, normalize=False):
        obj = self.copy()
        obj.multiply(val)

        return obj


class ComponentList(object):
    def __init__(self, members=None):
        self.members = members or []

    def append(self, member):
        self.members.append(member)

    @property
    def details(self):
        summary = defaultdict(float)

        for member in self.members:
            summary[member.name] += member.units

        return dict(summary)

    def sum(self):
        attrs = defaultdict(float)

        for member in self.members:
            attrs["units"] += member.units

            for k, v in member.attrs.items():
                if k not in attrs:
                    attrs[k] = 0

                attrs[k] += v

        return dict(attrs)
"""
